# Copyright 2019 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import copy
import uuid

from oslo_utils import timeutils
from ovs.db import idl as ovs_idl
from ovsdbapp.backend.ovs_idl import command
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp.backend.ovs_idl import rowview
from ovsdbapp.schema.ovn_northbound import commands as ovn_nb_commands
from ovsdbapp import utils as ovsdbapp_utils

from neutron._i18n import _
from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import exceptions as ovn_exc
from neutron.common.ovn import utils
from neutron.services.portforwarding.constants import PORT_FORWARDING_PREFIX

from oslo_log import log


LOG = log.getLogger(__name__)

RESOURCE_TYPE_MAP = {
    ovn_const.TYPE_NETWORKS: 'Logical_Switch',
    ovn_const.TYPE_PORTS: 'Logical_Switch_Port',
    ovn_const.TYPE_ROUTERS: 'Logical_Router',
    ovn_const.TYPE_ROUTER_PORTS: 'Logical_Router_Port',
    ovn_const.TYPE_FLOATINGIPS: 'NAT',
    ovn_const.TYPE_SUBNETS: 'DHCP_Options',
    ovn_const.TYPE_ADDRESS_GROUPS: 'Address_Set',
}


def _addvalue_to_list(row, column, new_value):
    row.addvalue(column, new_value)


def _delvalue_from_list(row, column, old_value):
    row.delvalue(column, old_value)


def _updatevalues_in_list(row, column, new_values=None, old_values=None):
    new_values = new_values or []
    old_values = old_values or []

    for new_value in new_values:
        row.addvalue(column, new_value)
    for old_value in old_values:
        row.delvalue(column, old_value)


def get_lsp_dhcp_options_uuids(lsp, lsp_name):
    # Get dhcpv4_options and dhcpv6_options uuids from Logical_Switch_Port,
    # which are references of port dhcp options in DHCP_Options table.
    uuids = set()
    for dhcp_opts in getattr(lsp, 'dhcpv4_options', []):
        external_ids = getattr(dhcp_opts, 'external_ids', {})
        if external_ids.get('port_id') == lsp_name:
            uuids.add(dhcp_opts.uuid)
    for dhcp_opts in getattr(lsp, 'dhcpv6_options', []):
        external_ids = getattr(dhcp_opts, 'external_ids', {})
        if external_ids.get('port_id') == lsp_name:
            uuids.add(dhcp_opts.uuid)
    return uuids


def _add_gateway_chassis(api, txn, lrp_name, val):
    gateway_chassis = api._tables.get('Gateway_Chassis')
    if not gateway_chassis:
        chassis = {ovn_const.OVN_GATEWAY_CHASSIS_KEY: val[0]}
        return 'options', chassis
    prio = len(val)
    uuid_list = []
    for chassis in val:
        gwc_name = f'{lrp_name}_{chassis}'
        try:
            gwc = idlutils.row_by_value(
                api.idl, 'Gateway_Chassis', 'name', gwc_name)
        except idlutils.RowNotFound:
            gwc = txn.insert(gateway_chassis)
            gwc.name = gwc_name
        gwc.chassis_name = chassis
        gwc.priority = prio
        LOG.info(
            "Schedule LRP %(lrp)s on gateway %(gtw)s with priority %(prio)s",
            {"lrp": lrp_name, "gtw": chassis, "prio": prio})
        prio = prio - 1
        uuid_list.append(gwc.uuid)
    return 'gateway_chassis', uuid_list


def _sync_ha_chassis_group(txn, nb_api, name, chassis_priority,
                           may_exist=False, table_name='HA_Chassis_Group',
                           **columns):
    result = None
    hcg = nb_api.lookup(table_name, name, default=None)
    if hcg:
        if not may_exist:
            raise RuntimeError(_('HA_Chassis_Group %s exists') % name)
    else:
        hcg = txn.insert(nb_api._tables[table_name])
        hcg.name = name
        command.BaseCommand.set_columns(hcg, **columns)
        result = hcg.uuid

    # HA_Chassis registers handling.
    # Remove the non-existing chassis in ``self.chassis_priority``
    hc_to_remove = []
    for hc in getattr(hcg, 'ha_chassis', []):
        if hc.chassis_name not in chassis_priority:
            hc_to_remove.append(hc)

    for hc in hc_to_remove:
        hcg.delvalue('ha_chassis', hc)
        hc.delete()

    # Update the priority of the existing chassis.
    for hc in getattr(hcg, 'ha_chassis', []):
        hc_priority = chassis_priority.pop(hc.chassis_name)
        hc.priority = hc_priority

    # Add the non-existing HA_Chassis registers.
    for hc_name, priority in chassis_priority.items():
        hc = txn.insert(nb_api.tables['HA_Chassis'])
        hc.chassis_name = hc_name
        hc.priority = priority
        hcg.addvalue('ha_chassis', hc)

    if not result:
        result = rowview.RowView(hcg)

    return result


class CheckLivenessCommand(command.BaseCommand):
    def run_idl(self, txn):
        # txn.pre_commit responsible for updating nb_global.nb_cfg, but
        # python-ovs will not update nb_cfg if no other changes are made
        self.api.nb_global.setkey('external_ids',
                                  ovn_const.OVN_LIVENESS_CHECK_EXT_ID_KEY,
                                  str(timeutils.utcnow(with_timezone=True)))
        self.result = self.api.nb_global.nb_cfg


class AddNetworkCommand(command.AddCommand):
    table_name = 'Logical_Switch'

    def __init__(self, api, network_id, may_exist=False, **columns):
        super().__init__(api)
        self.network_uuid = uuid.UUID(str(network_id))
        self.may_exist = may_exist
        self.columns = columns

    def run_idl(self, txn):
        table = self.api.tables[self.table_name]
        try:
            ls = table.rows[self.network_uuid]
            if self.may_exist:
                self.result = rowview.RowView(ls)
                return
            msg = _("Switch %s already exists") % self.network_uuid
            raise RuntimeError(msg)
        except KeyError:
            # Adding a new LS
            if utils.ovs_persist_uuid_supported(txn.idl):
                ls = txn.insert(table, new_uuid=self.network_uuid,
                                persist_uuid=True)
            else:
                ls = txn.insert(table)
        self.set_columns(ls, **self.columns)
        ls.name = utils.ovn_name(self.network_uuid)
        self.result = ls.uuid


class DelLogicalSwitchCommand(command.BaseCommand):
    def __init__(self, api, ls_name, if_exists):
        super().__init__(api)
        self.ls_name = ls_name
        self.if_exists = if_exists

    def run_idl(self, txn):
        try:
            ls = self.api.lookup('Logical_Switch', self.ls_name)
        except idlutils.RowNotFound as e:
            if self.if_exists:
                return
            msg = "Logical Switch %s does not exist" % self.ls_name
            raise RuntimeError(msg) from e

        # Delete the DNS record associated to this Neutron network.
        for dns_row in ls.dns_records:
            if dns_row.external_ids.get('ls_name') == self.ls_name:
                dns_row.delete()
                break

        # Delete the Logical_Switch register.
        ls.delete()

        # Delete the HA_Chassis_Group register associated, if exists.
        hcg = self.api.lookup('HA_Chassis_Group', self.ls_name, default=None)
        if hcg:
            hcg.delete()


class AddLSwitchPortCommand(command.BaseCommand):
    def __init__(self, api, lport, lswitch, may_exist, network_id=None,
                 **columns):
        super().__init__(api)
        self.lport = lport
        self.lswitch = lswitch
        self.may_exist = may_exist
        self.network_uuid = uuid.UUID(str(network_id)) if network_id else None
        self.columns = columns

    def run_idl(self, txn):
        try:
            # We must look in the local cache first, because the LS may have
            # been created as part of the current transaction. or in the case
            # of adding an LSP to a LS that was created before persist_uuid
            lswitch = idlutils.row_by_value(self.api.idl, 'Logical_Switch',
                                            'name', self.lswitch)
        except idlutils.RowNotFound:
            if self.network_uuid and utils.ovs_persist_uuid_supported(txn.idl):
                # Create a "fake" row with the right UUID so python-ovs creates
                # a transaction referencing the Row, even though we might not
                # have received the update for the row ourselves.
                lswitch = ovs_idl.Row(self.api.idl,
                                      self.api.tables['Logical_Switch'],
                                      uuid=self.network_uuid, data={})
            else:
                msg = _("Logical Switch %s does not exist") % self.lswitch
                raise RuntimeError(msg)
        if self.may_exist:
            port = idlutils.row_by_value(self.api.idl,
                                         'Logical_Switch_Port', 'name',
                                         self.lport, None)
            if port:
                self.result = port.uuid
                return

        port = txn.insert(self.api._tables['Logical_Switch_Port'])
        port.name = self.lport
        port.tag = self.columns.pop('tag', []) or []
        dhcpv4_options = self.columns.pop('dhcpv4_options', [])
        if isinstance(dhcpv4_options, list):
            port.dhcpv4_options = dhcpv4_options
        else:
            port.dhcpv4_options = [dhcpv4_options.result]
        dhcpv6_options = self.columns.pop('dhcpv6_options', [])
        if isinstance(dhcpv6_options, list):
            port.dhcpv6_options = dhcpv6_options
        else:
            port.dhcpv6_options = [dhcpv6_options.result]

        # NOTE(ralonsoh): HA chassis group is created by Neutron, there is no
        # need to create it in this command.
        ha_chassis_group = self.columns.pop('ha_chassis_group', None)
        if ha_chassis_group:
            hcg_uuid = ovsdbapp_utils.get_uuid(ha_chassis_group)
            port.ha_chassis_group = hcg_uuid

        for col, val in self.columns.items():
            setattr(port, col, val)
        # add the newly created port to existing lswitch
        _addvalue_to_list(lswitch, 'ports', port.uuid)
        self.result = port.uuid

    def post_commit(self, txn):
        self.result = txn.get_insert_uuid(self.result)


class SetLSwitchPortCommand(command.BaseCommand):
    def __init__(self, api, lport, external_ids_update, if_exists, **columns):
        super().__init__(api)
        self.lport = lport
        self.external_ids_update = external_ids_update
        self.columns = columns
        self.if_exists = if_exists

    def run_idl(self, txn):
        try:
            port = idlutils.row_by_value(self.api.idl, 'Logical_Switch_Port',
                                         'name', self.lport)
        except idlutils.RowNotFound:
            if self.if_exists:
                return
            msg = _("Logical Switch Port %s does not exist") % self.lport
            raise RuntimeError(msg)

        # Delete DHCP_Options records no longer referred by this port.
        # The table rows should be consistent for the same transaction.
        # After we get DHCP_Options rows uuids from port dhcpv4_options
        # and dhcpv6_options references, the rows shouldn't disappear for
        # this transaction before we delete it.
        cur_port_dhcp_opts = get_lsp_dhcp_options_uuids(
            port, self.lport)
        new_port_dhcp_opts = set()
        dhcpv4_options = self.columns.pop('dhcpv4_options', None)
        if dhcpv4_options is None:
            new_port_dhcp_opts.update([option.uuid for option in
                                       getattr(port, 'dhcpv4_options', [])])
        elif isinstance(dhcpv4_options, list):
            new_port_dhcp_opts.update(dhcpv4_options)
            port.dhcpv4_options = dhcpv4_options
        else:
            new_port_dhcp_opts.add(dhcpv4_options.result)
            port.dhcpv4_options = [dhcpv4_options.result]
        dhcpv6_options = self.columns.pop('dhcpv6_options', None)
        if dhcpv6_options is None:
            new_port_dhcp_opts.update([option.uuid for option in
                                       getattr(port, 'dhcpv6_options', [])])
        elif isinstance(dhcpv6_options, list):
            new_port_dhcp_opts.update(dhcpv6_options)
            port.dhcpv6_options = dhcpv6_options
        else:
            new_port_dhcp_opts.add(dhcpv6_options.result)
            port.dhcpv6_options = [dhcpv6_options.result]
        for uuid_ in cur_port_dhcp_opts - new_port_dhcp_opts:
            self.api._tables['DHCP_Options'].rows[uuid_].delete()

        external_ids_update = self.external_ids_update or {}
        external_ids = getattr(port, 'external_ids', {})
        for k, v in external_ids_update.items():
            external_ids[k] = v
        port.external_ids = external_ids

        # NOTE(ralonsoh): HA chassis group is created by Neutron, there is no
        # need to create it in this command. The register is also deleted when
        # the network to which the HA chassis group is associated is deleted.
        ha_chassis_group = self.columns.pop('ha_chassis_group', None)
        if ha_chassis_group:
            hcg_uuid = ovsdbapp_utils.get_uuid(ha_chassis_group)
            try:
                port_hcg_uuid = port.ha_chassis_group[0].uuid
            except IndexError:
                port_hcg_uuid = None
            if port_hcg_uuid != hcg_uuid:
                port.ha_chassis_group = hcg_uuid
        elif ha_chassis_group == []:
            port.ha_chassis_group = []

        for col, val in self.columns.items():
            setattr(port, col, val)


class UpdateLSwitchPortQosOptionsCommand(command.BaseCommand):
    def __init__(self, api, lport, if_exists, **qos):
        super().__init__(api)
        self.lport = lport
        self.if_exists = if_exists
        self.qos = qos

    def run_idl(self, txn):
        # NOTE(ralonsoh): this command can be called from inside a transaction
        # where the LSP is being created. If this is not the case, the value
        # provided in the Neutron port ID (== LSP.name).
        if isinstance(self.lport, command.BaseCommand):
            port_id = self.lport.result
        else:
            port_id = self.lport

        try:
            port = self.api.lookup('Logical_Switch_Port', port_id)
        except idlutils.RowNotFound:
            if self.if_exists:
                return
            raise RuntimeError(_('Logical Switch Port %s does not exist') %
                               port_id)

        # TODO(ralonsoh): add a check to only modify the QoS related keys:
        # qos_max_rate, qos_burst and qos_min_rate.
        for key, value in self.qos.items():
            if value is None:
                port.delkey('options', key)
            else:
                port.setkey('options', key, value)


class DelLSwitchPortCommand(command.BaseCommand):
    def __init__(self, api, lport, lswitch, if_exists):
        super().__init__(api)
        self.lport = lport
        self.lswitch = lswitch
        self.if_exists = if_exists

    def run_idl(self, txn):
        try:
            lport = idlutils.row_by_value(self.api.idl, 'Logical_Switch_Port',
                                          'name', self.lport)
            lswitch = idlutils.row_by_value(self.api.idl, 'Logical_Switch',
                                            'name', self.lswitch)
        except idlutils.RowNotFound:
            if self.if_exists:
                return
            msg = _("Port %s does not exist") % self.lport
            raise RuntimeError(msg)

        # Delete DHCP_Options records no longer referred by this port.
        cur_port_dhcp_opts = get_lsp_dhcp_options_uuids(
            lport, self.lport)
        for uuid_ in cur_port_dhcp_opts:
            self.api._tables['DHCP_Options'].rows[uuid_].delete()

        # Delete the HA_Chassis_Group associated to an external port.
        if (lport.type == ovn_const.LSP_TYPE_EXTERNAL and
                lport.ha_chassis_group):
            hcg = lport.ha_chassis_group[0]
            lport.delvalue('ha_chassis_group', hcg)
            if hcg.name == utils.ovn_extport_chassis_group_name(lport.name):
                hcg.delete()

        _delvalue_from_list(lswitch, 'ports', lport)
        self.api._tables['Logical_Switch_Port'].rows[lport.uuid].delete()


class UpdateLRouterCommand(command.BaseCommand):
    def __init__(self, api, name, if_exists, **columns):
        super().__init__(api)
        self.name = name
        self.columns = columns
        self.if_exists = if_exists

    def run_idl(self, txn):
        try:
            lrouter = idlutils.row_by_value(self.api.idl, 'Logical_Router',
                                            'name', self.name, None)
        except idlutils.RowNotFound:
            if self.if_exists:
                return
            msg = _("Logical Router %s does not exist") % self.name
            raise RuntimeError(msg)

        if lrouter:
            for col, val in self.columns.items():
                setattr(lrouter, col, val)
            return


class ScheduleUnhostedGatewaysCommand(command.BaseCommand):
    def __init__(self, nb_api, g_name, sb_api, plugin, port_physnets,
                 all_gw_chassis, chassis_with_physnets, chassis_with_azs):
        super().__init__(api=nb_api)
        self.g_name = g_name
        self.sb_api = sb_api
        self.scheduler = plugin.scheduler
        self.ovn_client = plugin._ovn_client
        self.port_physnets = port_physnets
        self.all_gw_chassis = all_gw_chassis
        self.chassis_with_physnets = chassis_with_physnets
        self.chassis_with_azs = chassis_with_azs

    def run_idl(self, txn):
        lrouter_port = self.api.lookup("Logical_Router_Port", self.g_name)
        physnet = self.port_physnets.get(
            self.g_name[len(ovn_const.LRP_PREFIX):])
        # Remove any invalid gateway chassis from the list, otherwise
        # we can have a situation where all existing_chassis are invalid
        existing_chassis = self.api.get_gateway_chassis_binding(self.g_name)
        primary = existing_chassis[0] if existing_chassis else None
        az_hints = self.api.get_gateway_chassis_az_hints(self.g_name)
        filtered_existing_chassis = (
            self.scheduler.filter_existing_chassis(
                gw_chassis=self.all_gw_chassis, physnet=physnet,
                chassis_physnets=self.chassis_with_physnets,
                existing_chassis=existing_chassis, az_hints=az_hints,
                chassis_with_azs=self.chassis_with_azs))
        if existing_chassis != filtered_existing_chassis:
            first_diff = None
            for i in range(len(filtered_existing_chassis)):
                if existing_chassis[i] != filtered_existing_chassis[i]:
                    first_diff = i
                    break
            if first_diff is not None:
                LOG.debug(
                    "A chassis for this gateway has been filtered. "
                    "Rebalancing priorities %s and lower", first_diff)
                filtered_existing_chassis = filtered_existing_chassis[
                    :max(first_diff, 1)]

        candidates = self.ovn_client.get_candidates_for_scheduling(
            physnet, cms=self.all_gw_chassis,
            chassis_physnets=self.chassis_with_physnets,
            availability_zone_hints=az_hints)
        chassis = self.scheduler.select(
            self.api, self.sb_api, self.g_name, candidates=candidates,
            existing_chassis=filtered_existing_chassis)
        if primary and primary != chassis[0]:
            if primary not in chassis:
                LOG.debug("Primary gateway chassis %(old)s "
                          "has been removed from the system. Moving "
                          "gateway %(gw)s to other chassis %(new)s.",
                          {'gw': self.g_name,
                           'old': primary,
                           'new': chassis[0]})
            else:
                LOG.debug("Gateway %s is hosted at %s.", self.g_name, primary)
                # NOTE(mjozefcz): It means scheduler moved primary chassis
                # to other gw based on scheduling method. But we don't
                # want network flap - so moving actual primary to be on
                # the top.
                index = chassis.index(primary)
                chassis[0], chassis[index] = chassis[index], chassis[0]
        setattr(
            lrouter_port,
            *_add_gateway_chassis(self.api, txn, self.g_name, chassis))


class ScheduleNewGatewayCommand(command.BaseCommand):
    def __init__(self, nb_api, g_name, sb_api, lrouter_name, plugin, physnet,
                 az_hints):
        super().__init__(nb_api)
        self.g_name = g_name
        self.sb_api = sb_api
        self.lrouter_name = lrouter_name
        self.ovn_client = plugin._ovn_client
        self.scheduler = plugin.scheduler
        self.physnet = physnet
        self.az_hints = az_hints

    def run_idl(self, txn):
        lrouter = self.api.lookup("Logical_Router", self.lrouter_name)
        lrouter_port = self.api.lookup("Logical_Router_Port", self.g_name)

        candidates = self.ovn_client.get_candidates_for_scheduling(
            self.physnet, availability_zone_hints=self.az_hints)
        chassis = self.scheduler.select(
            self.api, self.sb_api, self.g_name, candidates=candidates,
            target_lrouter=lrouter)
        if chassis:
            setattr(lrouter_port,
                    *_add_gateway_chassis(self.api, txn, self.g_name, chassis))


class LrDelCommand(ovn_nb_commands.LrDelCommand):

    def run_idl(self, txn):
        super().run_idl(txn)
        try:
            hcg = self.api.lookup('HA_Chassis_Group', self.router)
            hcg.delete()
        except idlutils.RowNotFound:
            pass


class AddLRouterPortCommand(command.BaseCommand):
    def __init__(self, api, name, lrouter, may_exist, **columns):
        super().__init__(api)
        self.name = name
        self.lrouter = lrouter
        self.may_exist = may_exist
        self.columns = columns

    def run_idl(self, txn):

        try:
            lrouter = idlutils.row_by_value(self.api.idl, 'Logical_Router',
                                            'name', self.lrouter)
        except idlutils.RowNotFound:
            msg = _("Logical Router %s does not exist") % self.lrouter
            raise RuntimeError(msg)
        try:
            idlutils.row_by_value(self.api.idl, 'Logical_Router_Port',
                                  'name', self.name)
            if self.may_exist:
                return
            # The LRP entry with certain name has already exist, raise an
            # exception to notice caller. It's caller's responsibility to
            # call UpdateLRouterPortCommand to get LRP entry processed
            # correctly.
            msg = _("Logical Router Port with name \"%s\" "
                    "already exists.") % self.name
            raise RuntimeError(msg)
        except idlutils.RowNotFound:
            lrouter_port = txn.insert(self.api._tables['Logical_Router_Port'])
            lrouter_port.name = self.name
            for col, val in self.columns.items():
                if col == 'gateway_chassis':
                    col, val = _add_gateway_chassis(self.api, txn, self.name,
                                                    val)
                self.set_column(lrouter_port, col, val)
            _addvalue_to_list(lrouter, 'ports', lrouter_port)
            self.result = lrouter_port.uuid


class UpdateLRouterPortCommand(command.BaseCommand):
    def __init__(self, api, name, if_exists, **columns):
        super().__init__(api)
        self.name = name
        self.columns = columns
        self.if_exists = if_exists

    def run_idl(self, txn):
        try:
            lrouter_port = idlutils.row_by_value(self.api.idl,
                                                 'Logical_Router_Port',
                                                 'name', self.name)
        except idlutils.RowNotFound:
            if self.if_exists:
                return
            msg = _("Logical Router Port %s does not exist") % self.name
            raise RuntimeError(msg)

        for col, val in self.columns.items():
            if col == 'gateway_chassis':
                col, val = _add_gateway_chassis(self.api, txn, self.name,
                                                val)
            self.set_column(lrouter_port, col, val)


class DelLRouterPortCommand(command.BaseCommand):
    def __init__(self, api, name, lrouter, if_exists):
        super().__init__(api)
        self.name = name
        self.lrouter = lrouter
        self.if_exists = if_exists

    def run_idl(self, txn):
        try:
            lrouter_port = idlutils.row_by_value(self.api.idl,
                                                 'Logical_Router_Port',
                                                 'name', self.name)
        except idlutils.RowNotFound:
            if self.if_exists:
                return
            msg = _("Logical Router Port %s does not exist") % self.name
            raise RuntimeError(msg)
        try:
            lrouter = idlutils.row_by_value(self.api.idl, 'Logical_Router',
                                            'name', self.lrouter)
        except idlutils.RowNotFound:
            msg = _("Logical Router %s does not exist") % self.lrouter
            raise RuntimeError(msg)

        _delvalue_from_list(lrouter, 'ports', lrouter_port)
        lrouter_port.delete()


class SetLRouterPortInLSwitchPortCommand(command.BaseCommand):
    def __init__(self, api, lswitch_port, lrouter_port, is_gw_port,
                 if_exists, lsp_address):
        super().__init__(api)
        self.lswitch_port = lswitch_port
        self.lrouter_port = lrouter_port
        self.is_gw_port = is_gw_port
        self.if_exists = if_exists
        self.lsp_address = lsp_address

    def run_idl(self, txn):
        try:
            port = idlutils.row_by_value(self.api.idl, 'Logical_Switch_Port',
                                         'name', self.lswitch_port)
        except idlutils.RowNotFound:
            if self.if_exists:
                return
            msg = _("Logical Switch Port %s does not "
                    "exist") % self.lswitch_port
            raise RuntimeError(msg)

        options = {'router-port': self.lrouter_port}
        if self.is_gw_port:
            options[ovn_const.OVN_GATEWAY_NAT_ADDRESSES_KEY] = 'router'
            options[ovn_const.OVN_ROUTER_PORT_EXCLUDE_LB_VIPS_GARP] = 'true'
        setattr(port, 'options', options)
        setattr(port, 'type', 'router')
        setattr(port, 'addresses', self.lsp_address)


class SetLRouterMacAgeLimitCommand(command.BaseCommand):
    def __init__(self, api, router, threshold):
        super().__init__(api)
        self.router = router
        self.threshold = str(threshold)  # Just in case an integer sneaks in

    def run_idl(self, txn):
        # Creating a Command object that iterates over the list of Routers
        # from inside a transaction avoids the issue of doing two
        # transactions: one for list_rows() and the other for setting the
        # values on routers, which would allow routers to be added and removed
        # between the two transactions.
        if self.router is None:
            routers = self.api.tables["Logical_Router"].rows.values()
        else:
            routers = [self.api.lookup("Logical_Router", self.router)]

        for router in routers:
            # It's not technically necessary to check the value before setting
            # it as python-ovs is smart enough to avoid sending operations to
            # the server that would result in no change. The overhead of
            # setkey() though is > than the overhead of checking the value here
            try:
                if (router.options.get(ovn_const.LR_OPTIONS_MAC_AGE_LIMIT) ==
                        self.threshold):
                    continue
            except AttributeError:
                # The Logical_Router is newly created in this txn and has no
                # "options" set yet, which the following setkey will rectify
                pass
            router.setkey("options", ovn_const.LR_OPTIONS_MAC_AGE_LIMIT,
                          self.threshold)


class AddACLCommand(command.BaseCommand):
    def __init__(self, api, lswitch, lport, **columns):
        super().__init__(api)
        self.lswitch = lswitch
        self.lport = lport
        self.columns = columns

    def run_idl(self, txn):
        try:
            lswitch = idlutils.row_by_value(self.api.idl, 'Logical_Switch',
                                            'name', self.lswitch)
        except idlutils.RowNotFound:
            msg = _("Logical Switch %s does not exist") % self.lswitch
            raise RuntimeError(msg)

        row = txn.insert(self.api._tables['ACL'])
        for col, val in self.columns.items():
            setattr(row, col, val)
        _addvalue_to_list(lswitch, 'acls', row.uuid)


class DelACLCommand(command.BaseCommand):
    def __init__(self, api, lswitch, lport, if_exists):
        super().__init__(api)
        self.lswitch = lswitch
        self.lport = lport
        self.if_exists = if_exists

    def run_idl(self, txn):
        try:
            lswitch = idlutils.row_by_value(self.api.idl, 'Logical_Switch',
                                            'name', self.lswitch)
        except idlutils.RowNotFound:
            if self.if_exists:
                return
            msg = _("Logical Switch %s does not exist") % self.lswitch
            raise RuntimeError(msg)

        acls_to_del = []
        acls = getattr(lswitch, 'acls', [])
        for acl in acls:
            ext_ids = getattr(acl, 'external_ids', {})
            if ext_ids.get('neutron:lport') == self.lport:
                acls_to_del.append(acl)
        for acl in acls_to_del:
            acl.delete()
        _updatevalues_in_list(lswitch, 'acls', old_values=acls_to_del)


class DelACLBySGruleIDCommand(command.BaseCommand):
    lookup_table = 'Port_Group'

    def __init__(self, api, sg_id, sg_rule_id, if_exists):
        super().__init__(api)
        self.sg_id = sg_id
        self.sg_rule_id = sg_rule_id
        self.if_exists = if_exists

    def run_idl(self, txn):
        pg_name = utils.ovn_port_group_name(self.sg_id)
        try:
            port_group = idlutils.row_by_value(
                self.api.idl, self.lookup_table, 'name', pg_name)
        except idlutils.RowNotFound:
            if self.if_exists:
                return
            msg = _('%(table)s %(name)s does not exist') % {
                'table': self.lookup_table, 'name': pg_name}
            raise RuntimeError(msg)

        acls_to_del = None
        acls = getattr(port_group, 'acls', [])
        for acl in acls:
            ext_ids = getattr(acl, 'external_ids', {})
            if (ext_ids.get(ovn_const.OVN_SG_RULE_EXT_ID_KEY) ==
                    self.sg_rule_id):
                acls_to_del = acl
                break
        if acls_to_del:
            acls_to_del.delete()
            _updatevalues_in_list(port_group, 'acls', old_values=[acls_to_del])


class AddStaticRouteCommand(command.BaseCommand):
    def __init__(self, api, lrouter, maintain_bfd=False, **columns):
        super().__init__(api)
        self.lrouter = lrouter
        self.maintain_bfd = maintain_bfd
        self.columns = columns

    def run_idl(self, txn):
        try:
            lrouter = idlutils.row_by_value(self.api.idl, 'Logical_Router',
                                            'name', self.lrouter)
        except idlutils.RowNotFound:
            msg = _("Logical Router %s does not exist") % self.lrouter
            raise RuntimeError(msg)

        bfd_uuid = None
        if (self.maintain_bfd and
                'nexthop' in self.columns and
                'output_port' in self.columns):
            cmd = ovn_nb_commands.BFDAddCommand(self.api,
                                                self.columns['output_port'],
                                                self.columns['nexthop'],
                                                may_exist=True)
            cmd.run_idl(txn)
            try:
                bfd_uuid = cmd.result.uuid
            except AttributeError:
                # When the BFD record is created in the same transaction the
                # post commit code that would resolve the real UUID and look up
                # the bfd record has not run yet, and consequently the object
                # returned by BFDAddCommand() is an UUID object.
                bfd_uuid = cmd.result

        row = txn.insert(self.api._tables['Logical_Router_Static_Route'])
        for col, val in self.columns.items():
            setattr(row, col, val)
        if bfd_uuid:
            setattr(row, 'bfd', bfd_uuid)
        _addvalue_to_list(lrouter, 'static_routes', row.uuid)


class DelStaticRoutesCommand(command.BaseCommand):
    def __init__(self, api, lrouter, routes, if_exists):
        super().__init__(api)
        self.lrouter = lrouter
        self.routes = routes
        self.if_exists = if_exists

    def run_idl(self, txn):
        try:
            lrouter = idlutils.row_by_value(self.api.idl, 'Logical_Router',
                                            'name', self.lrouter)
        except idlutils.RowNotFound:
            if self.if_exists:
                return
            msg = _("Logical Router %s does not exist") % self.lrouter
            raise RuntimeError(msg)

        routes_to_be_deleted = []
        for route in getattr(lrouter, 'static_routes', []):
            route_tuple = (getattr(route, 'ip_prefix', ''),
                           getattr(route, 'nexthop', ''))
            if route_tuple in self.routes:
                routes_to_be_deleted.append(route)

        for route in routes_to_be_deleted:
            _delvalue_from_list(lrouter, 'static_routes', route)
            route.delete()


class SetStaticRouteCommand(command.BaseCommand):
    def __init__(self, api, sroute, **columns):
        super().__init__(api)
        self.sroute = sroute
        self.columns = columns

    def run_idl(self, txn):
        try:
            for col, val in self.columns.items():
                setattr(self.sroute, col, val)

        except idlutils.RowNotFound:
            msg = (_('Logical Router Static Route %s does not exist')
                   % self.sroute)
            raise RuntimeError(msg)


class UpdateObjectExtIdsCommand(command.BaseCommand, metaclass=abc.ABCMeta):
    table: str
    field = 'name'

    def __init__(self, api, record, external_ids, if_exists):
        super().__init__(api)
        self.record = record
        self.external_ids = external_ids
        self.if_exists = if_exists

    def run_idl(self, txn):
        try:
            # api.lookup() would be better as it doesn't rely on hardcoded col
            obj = idlutils.row_by_value(self.api.idl, self.table, self.field,
                                        self.record)
        except idlutils.RowNotFound:
            if self.if_exists:
                return
            raise RuntimeError(
                _("%(table)s %(record)s does not exist. "
                  "Cannot update external IDs") %
                  {'table': self.table, 'record': self.record})

        for ext_id_key, ext_id_value in self.external_ids.items():
            obj.setkey('external_ids', ext_id_key, ext_id_value)


class UpdateChassisExtIdsCommand(UpdateObjectExtIdsCommand):
    table = 'Chassis'


class UpdatePortBindingExtIdsCommand(UpdateObjectExtIdsCommand):
    table = 'Port_Binding'
    field = 'logical_port'


class UpdateLbExternalIds(UpdateObjectExtIdsCommand):
    table = 'Load_Balancer'


class AddDHCPOptionsCommand(command.BaseCommand):
    def __init__(self, api, subnet_id, port_id=None, may_exist=True,
                 **columns):
        super().__init__(api)
        self.columns = columns
        self.may_exist = may_exist
        self.subnet_id = subnet_id
        self.port_id = port_id
        self.new_insert = False

    def _get_dhcp_options_row(self):
        for row in self.api._tables['DHCP_Options'].rows.values():
            external_ids = getattr(row, 'external_ids', {})
            port_id = external_ids.get('port_id')
            if self.subnet_id == external_ids.get('subnet_id'):
                if self.port_id == port_id:
                    return row

    def run_idl(self, txn):
        row = None
        if self.may_exist:
            row = self._get_dhcp_options_row()

        if not row:
            row = txn.insert(self.api._tables['DHCP_Options'])
            self.new_insert = True
        for col, val in self.columns.items():
            setattr(row, col, val)
        self.result = row.uuid

    def post_commit(self, txn):
        # Update the result with inserted uuid for new inserted row, or the
        # uuid get in run_idl should be real uuid already.
        if self.new_insert:
            self.result = txn.get_insert_uuid(self.result)


class DelDHCPOptionsCommand(command.BaseCommand):
    def __init__(self, api, row_uuid, if_exists=True):
        super().__init__(api)
        self.if_exists = if_exists
        self.row_uuid = row_uuid

    def run_idl(self, txn):
        if self.row_uuid not in self.api._tables['DHCP_Options'].rows:
            if self.if_exists:
                return
            msg = _("DHCP Options row %s does not exist") % self.row_uuid
            raise RuntimeError(msg)

        self.api._tables['DHCP_Options'].rows[self.row_uuid].delete()


class AddNATRuleInLRouterCommand(command.BaseCommand):
    # TODO(chandrav): Add unit tests, bug #1638715.
    def __init__(self, api, lrouter, **columns):
        super().__init__(api)
        self.lrouter = lrouter
        self.columns = columns

    def run_idl(self, txn):
        try:
            lrouter = idlutils.row_by_value(self.api.idl, 'Logical_Router',
                                            'name', self.lrouter)
        except idlutils.RowNotFound:
            msg = _("Logical Router %s does not exist") % self.lrouter
            raise RuntimeError(msg)

        row = txn.insert(self.api._tables['NAT'])
        for col, val in self.columns.items():
            setattr(row, col, val)
        lrouter.addvalue('nat', row.uuid)


class DeleteNATRuleInLRouterCommand(command.BaseCommand):
    # TODO(chandrav): Add unit tests, bug #1638715.
    def __init__(self, api, lrouter, type, logical_ip, external_ip,
                 if_exists):
        super().__init__(api)
        self.lrouter = lrouter
        self.type = type
        self.logical_ip = logical_ip
        self.external_ip = external_ip
        self.if_exists = if_exists

    def run_idl(self, txn):
        try:
            lrouter = idlutils.row_by_value(self.api.idl, 'Logical_Router',
                                            'name', self.lrouter)
        except idlutils.RowNotFound:
            if self.if_exists:
                return
            msg = _("Logical Router %s does not exist") % self.lrouter
            raise RuntimeError(msg)

        for nat in lrouter.nat:
            if (self.type == nat.type and
                    self.external_ip == nat.external_ip and
                    self.logical_ip == nat.logical_ip):
                lrouter.delvalue('nat', nat)
                nat.delete()
                break


class SetNATRuleInLRouterCommand(command.BaseCommand):
    def __init__(self, api, lrouter, nat_rule_uuid, **columns):
        super().__init__(api)
        self.lrouter = lrouter
        self.nat_rule_uuid = nat_rule_uuid
        self.columns = columns

    def run_idl(self, txn):
        try:
            lrouter = idlutils.row_by_value(self.api.idl, 'Logical_Router',
                                            'name', self.lrouter)
        except idlutils.RowNotFound:
            msg = _("Logical Router %s does not exist") % self.lrouter
            raise RuntimeError(msg)

        for nat_rule in lrouter.nat:
            if nat_rule.uuid == self.nat_rule_uuid:
                for col, val in self.columns.items():
                    setattr(nat_rule, col, val)
                break


class CheckRevisionNumberCommand(command.BaseCommand):

    def __init__(self, api, name, resource, resource_type, if_exists):
        super().__init__(api)
        self.name = name
        self.resource = resource
        self.resource_type = resource_type
        self.if_exists = if_exists

    def _get_floatingip_or_pf(self):
        # TYPE_FLOATINGIPS: Determine table to use based on name.
        # Floating ip port forwarding resources are kept in load
        # balancer table and have a well known name.
        if self.name.startswith(PORT_FORWARDING_PREFIX):
            return self.api.lookup('Load_Balancer', self.name)

        # TODO(lucasagomes): We can't use self.api.lookup() because that
        # method does not introspect map type columns. We could either:
        # 1. Enhance it to look into maps or, 2. Add a new ``name`` column
        # to the NAT table so that we can use lookup() just like we do
        # for other resources
        for nat in self.api._tables['NAT'].rows.values():
            if nat.type != 'dnat_and_snat':
                continue
            ext_ids = getattr(nat, 'external_ids', {})
            if ext_ids.get(ovn_const.OVN_FIP_EXT_ID_KEY) == self.name:
                return nat

        raise idlutils.RowNotFound(
            table='NAT', col='external_ids', match=self.name)

    def _get_subnet(self):
        for dhcp in self.api._tables['DHCP_Options'].rows.values():
            ext_ids = getattr(dhcp, 'external_ids', {})
            # Ignore ports DHCP Options
            if ext_ids.get('port_id'):
                continue
            if ext_ids.get('subnet_id') == self.name:
                return dhcp

        raise idlutils.RowNotFound(
            table='DHCP_Options', col='external_ids', match=self.name)

    def run_idl(self, txn):
        try:
            ovn_table = RESOURCE_TYPE_MAP[self.resource_type]
            ovn_resource = None
            if self.resource_type == ovn_const.TYPE_FLOATINGIPS:
                ovn_resource = self._get_floatingip_or_pf()
            elif self.resource_type == ovn_const.TYPE_SUBNETS:
                ovn_resource = self._get_subnet()
            else:
                ovn_resource = self.api.lookup(ovn_table, self.name)
        except idlutils.RowNotFound:
            if self.if_exists:
                return
            msg = (_('Failed to check the revision number for %s: Resource '
                     'does not exist') % self.name)
            raise RuntimeError(msg)

        external_ids = getattr(ovn_resource, 'external_ids', {})
        ovn_revision = int(external_ids.get(
            ovn_const.OVN_REV_NUM_EXT_ID_KEY, -1))
        neutron_revision = utils.get_revision_number(self.resource,
                                                     self.resource_type)
        if ovn_revision > neutron_revision:
            raise ovn_exc.RevisionConflict(
                resource_id=self.name, resource_type=self.resource_type)

        ovn_resource.verify('external_ids')
        ovn_resource.setkey('external_ids', ovn_const.OVN_REV_NUM_EXT_ID_KEY,
                            str(neutron_revision))

    def post_commit(self, txn):
        self.result = ovn_const.TXN_COMMITTED


class DeleteLRouterExtGwCommand(command.BaseCommand):

    def __init__(self, api, lrouter, if_exists, maintain_bfd=True):
        super().__init__(api)
        self.lrouter = lrouter
        self.if_exists = if_exists
        self.maintain_bfd = maintain_bfd

    def run_idl(self, txn):
        try:
            lrouter = idlutils.row_by_value(self.api.idl, 'Logical_Router',
                                            'name', self.lrouter)
        except idlutils.RowNotFound:
            if self.if_exists:
                return
            msg = _("Logical Router %s does not exist") % self.lrouter
            raise RuntimeError(msg)

        if self.maintain_bfd:
            lrp_names = set()
            for lrp in getattr(lrouter, 'ports', []):
                lrp_names.add(lrp.name)
        for route in lrouter.static_routes:
            external_ids = getattr(route, 'external_ids', {})
            if ovn_const.OVN_ROUTER_IS_EXT_GW in external_ids:
                bfd = getattr(route, 'bfd', [])
                if bfd and self.maintain_bfd:
                    for bfd_rec in bfd:
                        bfd_logical_port = getattr(bfd_rec, 'logical_port', '')
                        if bfd_logical_port in lrp_names:
                            route.delvalue('bfd', bfd_rec)
                            bfd_rec.delete()
                lrouter.delvalue('static_routes', route)
                route.delete()

        for nat in lrouter.nat:
            if nat.type != 'snat':
                continue
            lrouter.delvalue('nat', nat)
            nat.delete()

        # Remove the router pinning to a chassis (if any).
        lrouter.delkey('options', 'chassis')

        # Remove the HA_Chassis_Group of the router (if any).
        hcg = self.api.lookup('HA_Chassis_Group',
                              lrouter.name, default=None)
        if hcg:
            hcg.delete()

        for gw_port in self.api.get_lrouter_gw_ports(lrouter.name):
            lrouter.delvalue('ports', gw_port)


class SetLSwitchPortToVirtualTypeCommand(command.BaseCommand):
    def __init__(self, api, lport, vip, parent, if_exists):
        super().__init__(api)
        self.lport = lport
        self.vip = vip
        self.parent = parent
        self.if_exists = if_exists

    def run_idl(self, txn):
        try:
            lsp = idlutils.row_by_value(self.api.idl, 'Logical_Switch_Port',
                                        'name', self.lport)
        except idlutils.RowNotFound:
            if self.if_exists:
                return
            msg = _("Logical Switch Port %s does not exist") % self.lport
            raise RuntimeError(msg)

        options = lsp.options
        options[ovn_const.LSP_OPTIONS_VIRTUAL_IP_KEY] = self.vip
        virtual_parents = options.get(
            ovn_const.LSP_OPTIONS_VIRTUAL_PARENTS_KEY, set())
        if virtual_parents:
            virtual_parents = set(virtual_parents.split(','))

        virtual_parents.add(self.parent)
        options[ovn_const.LSP_OPTIONS_VIRTUAL_PARENTS_KEY] = ','.join(
            virtual_parents)
        setattr(lsp, 'options', options)
        setattr(lsp, 'type', ovn_const.LSP_TYPE_VIRTUAL)


class UnsetLSwitchPortToVirtualTypeCommand(command.BaseCommand):
    def __init__(self, api, lport, parent, if_exists):
        super().__init__(api)
        self.lport = lport
        self.parent = parent
        self.if_exists = if_exists

    def run_idl(self, txn):
        try:
            lsp = idlutils.row_by_value(self.api.idl, 'Logical_Switch_Port',
                                        'name', self.lport)
        except idlutils.RowNotFound:
            if self.if_exists:
                return
            msg = _("Logical Switch Port %s does not exist") % self.lport
            raise RuntimeError(msg)

        options = lsp.options
        virtual_parents = options.get(
            ovn_const.LSP_OPTIONS_VIRTUAL_PARENTS_KEY, set())
        if virtual_parents:
            virtual_parents = set(virtual_parents.split(','))

        try:
            virtual_parents.remove(self.parent)
        except KeyError:
            pass

        # If virtual-parents is now empty, change the type and remove the
        # virtual-parents and virtual-ip options
        if not virtual_parents:
            options.pop(ovn_const.LSP_OPTIONS_VIRTUAL_PARENTS_KEY, None)
            options.pop(ovn_const.LSP_OPTIONS_VIRTUAL_IP_KEY, None)
            setattr(lsp, 'type', '')
        else:
            options[ovn_const.LSP_OPTIONS_VIRTUAL_PARENTS_KEY] = ','.join(
                virtual_parents)

        setattr(lsp, 'options', options)


class HAChassisGroupWithHCAddCommand(command.AddCommand):
    table_name = 'HA_Chassis_Group'

    def __init__(self, api, name, chassis_priority, may_exist=False,
                 **columns):
        super().__init__(api)
        self.name = name
        self.chassis_priority = copy.deepcopy(chassis_priority)
        self.may_exist = may_exist
        self.columns = columns

    def run_idl(self, txn):
        # HA_Chassis_Group register creation.
        self.result = _sync_ha_chassis_group(
            txn, self.api, self.name, self.chassis_priority,
            may_exist=self.may_exist, table_name=self.table_name,
            **self.columns)



# Copyright (c) 2023 China Unicom Cloud Data Co.,Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from neutron_lib.plugins.ml2 import ovs_constants as p_const

from os_ken.lib.packet import ether_types
from os_ken.lib.packet import in_proto

from neutron.agent.l2.extensions.metadata import metadata_flows_process
from neutron.agent.l2.extensions.metadata import metadata_path
from neutron.tests.unit.plugins.ml2.drivers.openvswitch.agent.openflow.native \
    import ovs_bridge_test_base

call = mock.call  # short hand


class MetadataDataPathFlowsTestCase(ovs_bridge_test_base.OVSBridgeTestMixin):

    def setUp(self):
        super().setUp()
        self.int_br = self.mock_bridge_cls('br-int', self.br_int_cls)
        self.path_br = self.mock_bridge_cls('br-phys', self.br_phys_cls)
        self.flows = metadata_flows_process.MetadataDataPathFlows()
        self.flows.set_path_br(self.path_br)
        self.flows.int_br = self.int_br
        self.provider_gateway_ip = "1.1.1.1"
        self.metadata_host_info = {
            "gateway_ip": self.provider_gateway_ip,
            "provider_ip": self.provider_gateway_ip,
            "mac_address": metadata_path.DEFAULT_META_GATEWAY_MAC,
            "service_protocol_port": 55555}

        mock__send_msg = mock.patch.object(self.int_br, '_send_msg').start()
        mock_delete_flows = mock.patch.object(self.int_br,
                                              'uninstall_flows').start()
        self.int_br.mock = mock.Mock()
        self.int_br.mock.attach_mock(mock__send_msg, '_send_msg')
        self.int_br.mock.attach_mock(mock_delete_flows, 'uninstall_flows')

    def test_metadata_path_defaults(self):
        provider_vlan = 1000
        ofport_patch_meta_to_int = 100
        ofport_tap_meta = 200
        self.flows.metadata_path_defaults(
            provider_vlan, ofport_patch_meta_to_int,
            ofport_tap_meta, self.metadata_host_info)

        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[],
                match=ofpp.OFPMatch(),
                priority=0,
                table_id=p_const.METADATA_EGRESS_NAT),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[],
                match=ofpp.OFPMatch(),
                priority=0,
                table_id=p_const.METADATA_IP_ARP_RESPONDER),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[],
                match=ofpp.OFPMatch(),
                priority=0,
                table_id=p_const.METADATA_INGRESS_DIRECT),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[
                    ofpp.OFPInstructionGotoTable(
                        table_id=p_const.METADATA_EGRESS_NAT),
                ],
                match=ofpp.OFPMatch(
                    in_port=ofport_patch_meta_to_int,
                    ipv4_dst=metadata_flows_process.METADATA_V4_IP,
                    eth_type=ether_types.ETH_TYPE_IP),
                priority=201,
                table_id=p_const.LOCAL_SWITCHING),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[
                    ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                        ofpp.OFPActionPushVlan(),
                        ofpp.OFPActionSetField(
                            vlan_vid=provider_vlan | ofp.OFPVID_PRESENT),
                    ]),
                    ofpp.OFPInstructionGotoTable(
                        table_id=p_const.METADATA_INGRESS_DIRECT),
                ],
                match=ofpp.OFPMatch(
                    in_port=ofport_tap_meta,
                    eth_type=ether_types.ETH_TYPE_IP),
                priority=201,
                table_id=p_const.LOCAL_SWITCHING),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[
                    ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                        ofpp.OFPActionSetField(
                            eth_dst=metadata_path.DEFAULT_META_GATEWAY_MAC),
                        ofpp.OFPActionSetField(
                            ipv4_dst=self.provider_gateway_ip),
                        ofpp.OFPActionSetField(tcp_dst=55555),
                        ofpp.OFPActionOutput(ofport_tap_meta, 0),
                    ]),
                ],
                match=ofpp.OFPMatch(
                    eth_type=ether_types.ETH_TYPE_IP,
                    ipv4_dst=metadata_flows_process.METADATA_V4_IP,
                    tcp_dst=80,
                    ip_proto=in_proto.IPPROTO_TCP),
                priority=202,
                table_id=p_const.METADATA_EGRESS_OUTPUT),
                           active_bundle=None),
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[
                    ofpp.OFPInstructionGotoTable(
                        table_id=p_const.METADATA_IP_ARP_RESPONDER),
                ],
                match=ofpp.OFPMatch(
                    in_port=ofport_tap_meta,
                    eth_type=ether_types.ETH_TYPE_ARP),
                priority=201,
                table_id=p_const.LOCAL_SWITCHING),
                           active_bundle=None),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_metadata_egress_direct(self):
        ofport_patch_meta_to_int = 100
        self.flows.metadata_egress_direct(ofport_patch_meta_to_int)

        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[
                    ofpp.OFPInstructionGotoTable(
                        table_id=p_const.METADATA_EGRESS_NAT),
                ],
                match=ofpp.OFPMatch(
                    in_port=ofport_patch_meta_to_int,
                    ipv4_dst=metadata_flows_process.METADATA_V4_IP,
                    eth_type=ether_types.ETH_TYPE_IP),
                priority=201,
                table_id=p_const.LOCAL_SWITCHING),
                           active_bundle=None),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_metadata_ingress_direct(self):
        provider_vlan = 1000
        ofport_tap_meta = 200
        self.flows.metadata_ingress_direct(
            provider_vlan, ofport_tap_meta)

        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[
                    ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                        ofpp.OFPActionPushVlan(),
                        ofpp.OFPActionSetField(
                            vlan_vid=provider_vlan | ofp.OFPVID_PRESENT),
                    ]),
                    ofpp.OFPInstructionGotoTable(
                        table_id=p_const.METADATA_INGRESS_DIRECT),
                ],
                match=ofpp.OFPMatch(
                    in_port=ofport_tap_meta,
                    eth_type=ether_types.ETH_TYPE_IP),
                priority=201,
                table_id=p_const.LOCAL_SWITCHING),
                           active_bundle=None),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_metadata_arp_direct(self):
        ofport_tap_meta = 200
        self.flows.metadata_arp_direct(ofport_tap_meta)

        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[
                    ofpp.OFPInstructionGotoTable(
                        table_id=p_const.METADATA_IP_ARP_RESPONDER),
                ],
                match=ofpp.OFPMatch(
                    in_port=ofport_tap_meta,
                    eth_type=ether_types.ETH_TYPE_ARP),
                priority=201,
                table_id=p_const.LOCAL_SWITCHING),
                           active_bundle=None),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_metadata_path_classify(self):
        ofport_tap_meta = 200
        self.flows.metadata_path_classify(
            ofport_tap_meta, self.metadata_host_info)

        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[
                    ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                        ofpp.OFPActionSetField(
                            eth_dst=metadata_path.DEFAULT_META_GATEWAY_MAC),
                        ofpp.OFPActionSetField(
                            ipv4_dst=self.provider_gateway_ip),
                        ofpp.OFPActionSetField(tcp_dst=55555),
                        ofpp.OFPActionOutput(ofport_tap_meta, 0),
                    ]),
                ],
                match=ofpp.OFPMatch(
                    eth_type=ether_types.ETH_TYPE_IP,
                    ipv4_dst=metadata_flows_process.METADATA_V4_IP,
                    tcp_dst=80,
                    ip_proto=in_proto.IPPROTO_TCP),
                priority=202,
                table_id=p_const.METADATA_EGRESS_OUTPUT),
                           active_bundle=None),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_install_arp_responder_dhcp_port(self):
        dhcp_port_ip = "192.168.100.2"
        self.flows.install_arp_responder(
            ip=dhcp_port_ip,
            mac=metadata_path.METADATA_DEFAULT_MAC,
            bridge=self.int_br,
            table=p_const.TRANSIENT_TABLE)

        (dp, ofp, ofpp) = self.int_br._get_dp()
        expected = [
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.int_br.default_cookie,
                instructions=[
                    ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                        ofpp.OFPActionSetField(arp_op=2),
                        ofpp.NXActionRegMove(
                            dst_field='arp_tha',
                            n_bits=48,
                            src_field='arp_sha'),
                        ofpp.NXActionRegMove(
                            dst_field='arp_tpa',
                            n_bits=32,
                            src_field='arp_spa'),
                        ofpp.OFPActionSetField(
                            arp_sha=metadata_path.METADATA_DEFAULT_MAC),
                        ofpp.OFPActionSetField(arp_spa=dhcp_port_ip),
                        ofpp.NXActionRegMove(src_field='eth_src',
                                             dst_field='eth_dst',
                                             n_bits=48),
                        ofpp.OFPActionSetField(
                            eth_src=metadata_path.METADATA_DEFAULT_MAC),
                        ofpp.OFPActionOutput(ofp.OFPP_IN_PORT, 0),
                    ]),
                ],
                match=self.flows.get_arp_responder_match(
                    ofp, ofpp, ip=dhcp_port_ip),
                priority=200,
                table_id=p_const.TRANSIENT_TABLE),
                           active_bundle=None),
        ]
        self.assertEqual(expected, self.int_br.mock.mock_calls)

    def test_install_arp_responder_metadata_provider_ip(self):
        provider_ip = "100.100.100.100"
        provider_mac = "fa:16:ee:11:22:33"
        self.flows.install_arp_responder(provider_ip, provider_mac)

        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[
                    ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                        ofpp.OFPActionSetField(arp_op=2),
                        ofpp.NXActionRegMove(
                            dst_field='arp_tha',
                            n_bits=48,
                            src_field='arp_sha'),
                        ofpp.NXActionRegMove(
                            dst_field='arp_tpa',
                            n_bits=32,
                            src_field='arp_spa'),
                        ofpp.OFPActionSetField(arp_sha=provider_mac),
                        ofpp.OFPActionSetField(arp_spa=provider_ip),
                        ofpp.NXActionRegMove(src_field='eth_src',
                                             dst_field='eth_dst',
                                             n_bits=48),
                        ofpp.OFPActionSetField(eth_src=provider_mac),
                        ofpp.OFPActionOutput(ofp.OFPP_IN_PORT, 0),
                    ]),
                ],
                match=self.flows.get_arp_responder_match(
                    ofp, ofpp, ip=provider_ip),
                priority=200,
                table_id=p_const.METADATA_IP_ARP_RESPONDER),
                           active_bundle=None),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_delete_arp_responder_dhcp_port(self):
        vm_ofport = 10
        dhcp_port_ip = "192.168.100.2"
        self.flows.delete_arp_responder(
            bridge=self.int_br,
            ip=dhcp_port_ip,
            table=p_const.ARP_SPOOF_TABLE,
            in_port=vm_ofport)

        (dp, ofp, ofpp) = self.int_br._get_dp()
        expected = [
            call.uninstall_flows(
                table_id=p_const.ARP_SPOOF_TABLE,
                match=self.flows.get_arp_responder_match(
                    ofp, ofpp, ip=dhcp_port_ip, in_port=vm_ofport)),
        ]
        self.assertEqual(expected, self.int_br.mock.mock_calls)

    def test_add_flow_snat_br_meta(self):
        vm_local_vlan = 1
        vm_mac = "aa:aa:aa:aa:aa:aa"
        vm_ip = "192.168.100.100"
        provider_mac = "fa:16:ee:11:22:33"
        provider_ip = "100.100.100.100"
        self.flows.add_flow_snat_br_meta(
            vm_local_vlan, vm_mac, vm_ip,
            provider_mac, provider_ip)

        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[
                    ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                        ofpp.OFPActionPopVlan(),
                        ofpp.OFPActionSetField(eth_src=provider_mac),
                        ofpp.OFPActionSetField(ipv4_src=provider_ip),
                    ]),
                    ofpp.OFPInstructionGotoTable(
                        table_id=p_const.METADATA_EGRESS_OUTPUT)
                ],
                match=ofpp.OFPMatch(
                    vlan_vid=vm_local_vlan | ofp.OFPVID_PRESENT,
                    eth_type=ether_types.ETH_TYPE_IP,
                    eth_src=vm_mac,
                    ipv4_src=vm_ip),
                priority=201,
                table_id=p_const.METADATA_EGRESS_NAT),
                           active_bundle=None),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_add_flow_ingress_dnat_direct_to_int_br(self):
        vm_local_vlan = 1
        provider_vlan = 1000
        provider_ip_addr = "100.100.100.100"
        vm_mac = "aa:aa:aa:aa:aa:aa"
        vm_ip = "192.168.100.100"
        ofport_patch_meta_to_int = 100
        ofport_tap_meta = 200
        self.flows.add_flow_ingress_dnat_direct_to_int_br(
            vm_local_vlan, provider_vlan, provider_ip_addr,
            vm_mac, vm_ip,
            ofport_patch_meta_to_int, ofport_tap_meta)

        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.stamp,
                instructions=[
                    ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                        ofpp.OFPActionSetField(
                            vlan_vid=vm_local_vlan | ofp.OFPVID_PRESENT),
                        ofpp.OFPActionSetField(eth_dst=vm_mac),
                        ofpp.OFPActionSetField(
                            ipv4_src=metadata_flows_process.METADATA_V4_IP),
                        ofpp.OFPActionSetField(ipv4_dst=vm_ip),
                        ofpp.OFPActionSetField(
                            tcp_src=metadata_flows_process.METADATA_V4_PORT),
                        ofpp.OFPActionOutput(ofport_patch_meta_to_int, 0),
                    ]),
                ],
                match=ofpp.OFPMatch(
                    vlan_vid=provider_vlan | ofp.OFPVID_PRESENT,
                    in_port=ofport_tap_meta,
                    eth_type=ether_types.ETH_TYPE_IP,
                    ipv4_dst=provider_ip_addr,
                    ip_proto=in_proto.IPPROTO_TCP),
                priority=202,
                table_id=p_const.METADATA_INGRESS_DIRECT),
                           active_bundle=None),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_add_flow_int_br_egress_direct(self):
        vm_ofport = 10
        vm_local_vlan = 1
        patch_to_br_meta = 1
        self.flows.add_flow_int_br_egress_direct(
            vm_ofport, vm_local_vlan, patch_to_br_meta)

        (dp, ofp, ofpp) = self.int_br._get_dp()
        expected = [
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.int_br.default_cookie,
                instructions=[
                    ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                        ofpp.OFPActionPushVlan(),
                        ofpp.OFPActionSetField(
                            vlan_vid=vm_local_vlan | ofp.OFPVID_PRESENT),
                        ofpp.OFPActionOutput(patch_to_br_meta, 0),
                    ]),
                ],
                match=ofpp.OFPMatch(
                    in_port=vm_ofport,
                    eth_type=ether_types.ETH_TYPE_IP,
                    ipv4_dst=metadata_flows_process.METADATA_V4_IP),
                priority=200,
                table_id=p_const.LOCAL_SWITCHING),
                           active_bundle=None),
        ]
        self.assertEqual(expected, self.int_br.mock.mock_calls)

    def test_add_flow_int_br_ingress_output(self):
        ofport_from_br_meta = 1
        vm_local_vlan = 1
        vm_mac = "aa:aa:aa:aa:aa:aa"
        vm_ofport = 10
        self.flows.add_flow_int_br_ingress_output(
            ofport_from_br_meta, vm_local_vlan, vm_mac, vm_ofport)

        (dp, ofp, ofpp) = self.int_br._get_dp()
        expected = [
            call._send_msg(ofpp.OFPFlowMod(dp,
                cookie=self.int_br.default_cookie,
                instructions=[
                    ofpp.OFPInstructionActions(ofp.OFPIT_APPLY_ACTIONS, [
                        ofpp.OFPActionPopVlan(),
                        ofpp.OFPActionOutput(vm_ofport, 0),
                    ]),
                ],
                match=ofpp.OFPMatch(
                    in_port=ofport_from_br_meta,
                    eth_type=ether_types.ETH_TYPE_IP,
                    vlan_vid=vm_local_vlan | ofp.OFPVID_PRESENT,
                    eth_dst=vm_mac,
                    ipv4_src=metadata_flows_process.METADATA_V4_IP),
                priority=200,
                table_id=p_const.LOCAL_SWITCHING),
                           active_bundle=None),
        ]
        self.assertEqual(expected, self.int_br.mock.mock_calls)

    def test_remove_port_metadata_path_nat_and_arp_flow(self):
        port_vlan = 1
        vm_mac = "aa:aa:aa:aa:aa:aa"
        vm_ip = "192.168.100.100"
        provider_vlan = 1000
        provider_ip_addr = "100.100.100.100"
        self.flows.remove_port_metadata_path_nat_and_arp_flow(
            port_vlan, vm_mac, vm_ip, provider_vlan,
            provider_ip_addr)

        (dp, ofp, ofpp) = self._get_dp()
        expected = [
            call.uninstall_flows(
                table_id=p_const.METADATA_EGRESS_NAT,
                eth_type=ether_types.ETH_TYPE_IP,
                eth_src=vm_mac,
                ipv4_src=vm_ip,
                vlan_vid=port_vlan | ofp.OFPVID_PRESENT),
            call.uninstall_flows(
                table_id=p_const.METADATA_IP_ARP_RESPONDER,
                eth_type=ether_types.ETH_TYPE_ARP,
                arp_tpa=provider_ip_addr),
            call.uninstall_flows(
                table_id=p_const.METADATA_INGRESS_DIRECT,
                eth_type=ether_types.ETH_TYPE_IP,
                ipv4_dst=provider_ip_addr,
                vlan_vid=provider_vlan | ofp.OFPVID_PRESENT),
        ]
        self.assertEqual(expected, self.mock.mock_calls)

    def test_remove_port_metadata_direct_flow(self):
        vm_ofport = 10
        vm_local_vlan = 1
        vm_mac = "aa:aa:aa:aa:aa:aa"
        patch_to_br_meta = 1
        self.flows.remove_port_metadata_direct_flow(
            vm_ofport, vm_local_vlan, vm_mac,
            patch_to_br_meta)

        (dp, ofp, ofpp) = self.int_br._get_dp()
        expected = [
            call.uninstall_flows(
                table_id=p_const.LOCAL_SWITCHING,
                eth_type=ether_types.ETH_TYPE_IP,
                ipv4_dst=metadata_flows_process.METADATA_V4_IP,
                in_port=vm_ofport),
            call.uninstall_flows(
                table_id=p_const.LOCAL_SWITCHING,
                eth_type=ether_types.ETH_TYPE_IP,
                ipv4_src=metadata_flows_process.METADATA_V4_IP,
                in_port=patch_to_br_meta,
                vlan_vid=vm_local_vlan | ofp.OFPVID_PRESENT,
                eth_dst=vm_mac),
        ]
        self.assertEqual(expected, self.int_br.mock.mock_calls)



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from ovsdbapp.backend.ovs_idl import command as ovs_cmd

from neutron.services.bgp import constants


class SetChassisBgpBridgesCommand(ovs_cmd.BaseCommand):
    def __init__(self, api, chassis_name, bridge_name_list):
        super().__init__(api)
        self.chassis_name = chassis_name
        self.bridge_name_list = bridge_name_list

    def run_idl(self, txn):
        chassis = self.api.lookup('Chassis_Private', self.chassis_name)
        external_ids = chassis.external_ids
        external_ids[constants.CHASSIS_BGP_BRIDGES_EXT_ID_KEY] = ','.join(
            sorted(self.bridge_name_list))

        self.set_column(chassis, 'external_ids', external_ids)



# Copyright 2025 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

BGP_ROUTER_REDISTRIBUTE = 'connected-as-host,nat'
OVN_DYNAMIC_ROUTING_PORT_MAPPING = 'dynamic-routing-port-mapping'

LRP_OPTIONS_PORT_NAME = 'dynamic-routing-port-name'
LRP_OPTIONS_DYNAMIC_ROUTING_MAINTAIN_VRF = 'dynamic-routing-maintain-vrf'
LRP_NETWORK_NAME_EXT_ID_KEY = 'neutron-bgp-network-name'

LR_OPTIONS_DYNAMIC_ROUTING = 'dynamic-routing'
LR_OPTIONS_DYNAMIC_ROUTING_REDISTRIBUTE = 'dynamic-routing-redistribute'
LR_OPTIONS_DYNAMIC_ROUTING_VRF_ID = 'dynamic-routing-vrf-id'

HA_CHASSIS_GROUP_PRIORITY = 10

CHASSIS_BGP_BRIDGES_EXT_ID_KEY = 'neutron-bgp-bridges'

AGENT_BGP_PEER_BRIDGES = 'neutron-bgp-peer-bridges'
AGENT_BGP_EXT_NAME = 'ovn-bgp'

BGP_BRIDGE_NIC_TYPES = ('', 'system')

BGP_LRP_TO_CHASSIS = 'neutron-bgp-lrp-to-chassis-router'

RELATED_RESOURCE_TAG = 'neutron-bgp-related-resource'



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc
import io
import socketserver
from urllib import parse

import jinja2
from neutron_lib import constants
from oslo_log import log as logging
from oslo_utils import encodeutils
import requests
import webob

from neutron._i18n import _
from neutron.agent.metadata import proxy_base
from neutron.common import ipv6_utils
from neutron.common import utils as common_utils


LOG = logging.getLogger(__name__)


PROXY_SERVICE_NAME = 'haproxy'
PROXY_SERVICE_CMD = 'haproxy'

CONTENT_ENCODERS = {
    'gzip': b'\x1f\x8b\x08',
    'deflate': b'\x1f\x8b\x08'
}


class InvalidUserOrGroupException(Exception):
    pass


METADATA_HAPROXY_GLOBAL = """
global
    log         /dev/log local0 %(log_level)s
    log-tag     %(log_tag)s
    user        %(user)s
    group       %(group)s
    maxconn     1024
    nbthread    2
    pidfile     %(pidfile)s
    daemon

defaults
    log global
    mode http
    option httplog
    option dontlognull
    option http-server-close
    option forwardfor
    retries                 3
    timeout http-request    30s
    timeout connect         30s
    timeout client          32s
    timeout server          32s
    timeout http-keep-alive 30s
"""

RATE_LIMITED_CONFIG_TEMPLATE = """
backend base_rate_limiter
    stick-table type %(ip_version)s size 10k expire %(stick_table_expire)ss store http_req_rate(%(base_window_duration)ss)

backend burst_rate_limiter
    stick-table type %(ip_version)s size 10k expire %(stick_table_expire)ss store http_req_rate(%(burst_window_duration)ss)

listen listener
    bind %(host)s:%(port)s
    %(bind_v6_line)s

    http-request track-sc0 src table base_rate_limiter
    http-request track-sc1 src table burst_rate_limiter
    http-request deny deny_status 429 if { src_http_req_rate(base_rate_limiter) gt %(base_query_rate_limit)s }
    http-request deny deny_status 429 if { src_http_req_rate(burst_rate_limiter) gt %(burst_query_rate_limit)s }

    server metadata %(unix_socket_path)s
"""  # noqa: E501 line-length

RESPONSE = jinja2.Template("""HTTP/1.1 {{ http_code }}
Content-Type: text/plain; charset=UTF-8
Connection: close
Content-Length: {{ len }}

<html>
 <head>
  <title>{{ title }}</title>
 </head>
 <body>
  <h1>{{ body_title }}</h1>
  {{ body }}<br /><br />


 </body>
</html>""")
RESPONSE_LENGTH = 95


def parse_ip_versions(ip_versions):
    if not set(ip_versions).issubset({constants.IP_VERSION_4,
                                      constants.IP_VERSION_6}):
        LOG.warning('Invalid metadata address IP versions: %s. Metadata rate '
                    'limiting will not be enabled.', ip_versions)
        return
    if len(ip_versions) != 1:
        LOG.warning('Invalid metadata address IP versions: %s. Metadata rate '
                    'limiting cannot be enabled for IPv4 and IPv6 at the same '
                    'time. Metadata rate limiting will not be enabled.',
                    ip_versions)
        return
    return ip_versions[0]


def get_haproxy_config(cfg_info, rate_limiting_config, header_config_template,
                       unlimited_config_template):
    ip_version = parse_ip_versions(rate_limiting_config.ip_versions)
    if rate_limiting_config.rate_limit_enabled and ip_version:
        cfg_info['ip_version'] = (
            'ipv6' if ip_version == 6 else 'ip')
        cfg_info['base_window_duration'] = (
            rate_limiting_config['base_window_duration'])
        cfg_info['base_query_rate_limit'] = (
            rate_limiting_config['base_query_rate_limit'])
        cfg_info['burst_window_duration'] = (
            rate_limiting_config['burst_window_duration'])
        cfg_info['burst_query_rate_limit'] = (
            rate_limiting_config['burst_query_rate_limit'])
        cfg_info['stick_table_expire'] = max(
            rate_limiting_config['base_window_duration'],
            rate_limiting_config['burst_window_duration'])
        FINAL_CONFIG_TEMPLATE = (METADATA_HAPROXY_GLOBAL +
                                 RATE_LIMITED_CONFIG_TEMPLATE +
                                 header_config_template)
    else:
        FINAL_CONFIG_TEMPLATE = (METADATA_HAPROXY_GLOBAL +
                                 unlimited_config_template +
                                 header_config_template)

    return FINAL_CONFIG_TEMPLATE % cfg_info


def encode_http_reponse(http_code, title, message):
    """Return an encoded HTTP, providing the HTTP code, title and message"""
    length = RESPONSE_LENGTH + len(title) * 2 + len(message)
    reponse = RESPONSE.render(http_code=http_code, title=title,
                              body_title=title, body=message, len=length)
    return encodeutils.to_utf8(reponse)


class MetadataProxyHandlerBaseSocketServer(
        proxy_base.MetadataProxyHandlerBase,
        socketserver.StreamRequestHandler,
        metaclass=abc.ABCMeta):
    @staticmethod
    def _http_response(http_response, request):
        headerlist = list(http_response.headers.items())
        # We detect if content is compressed by magic signature,
        # when `content-encoding` is not present.
        if not http_response.headers.get('content-encoding'):
            if http_response.content[:3] == CONTENT_ENCODERS['gzip']:
                headerlist.append(('content-encoding', 'gzip'))

        _res = webob.Response(
            body=http_response.content,
            status=http_response.status_code,
            headerlist=headerlist)
        # The content of the response is decoded depending on the
        # "Context-Enconding" header, if present. The operation is limited to
        # ("gzip", "deflate"), as is in the ``webob.response.Response`` class.
        if _res.content_encoding in CONTENT_ENCODERS.keys():
            _res.decode_content()

        # NOTE(ralonsoh): there should be a better way to format the HTTP
        # response, adding the HTTP version to the ``webob.Response``
        # output string.
        out = request.http_version + ' ' + str(_res)
        if (int(_res.headers['content-length']) == 0 and
                _res.status_code == 200):
            # Add 2 extra \r\n to the result. HAProxy is also expecting
            # it even when the body is empty.
            out += '\r\n\r\n'
        return out.encode(http_response.encoding)

    def _proxy_request(self, instance_id, project_id, req):
        headers = {
            'X-Forwarded-For': req.headers.get('X-Forwarded-For'),
            'X-Instance-ID': instance_id,
            'X-Tenant-ID': project_id,
            'X-Instance-ID-Signature': common_utils.sign_instance_id(
                self.conf, instance_id)
        }

        nova_host_port = ipv6_utils.valid_ipv6_url(
            self.conf.nova_metadata_host,
            self.conf.nova_metadata_port)

        url = parse.urlunsplit((
            self.conf.nova_metadata_protocol,
            nova_host_port,
            req.path_info,
            req.query_string,
            ''))

        disable_ssl_certificate_validation = self.conf.nova_metadata_insecure
        if self.conf.auth_ca_cert and not disable_ssl_certificate_validation:
            verify_cert = self.conf.auth_ca_cert
        else:
            verify_cert = not disable_ssl_certificate_validation

        client_cert = None
        if self.conf.nova_client_cert and self.conf.nova_client_priv_key:
            client_cert = (self.conf.nova_client_cert,
                           self.conf.nova_client_priv_key)

        try:
            resp = requests.request(method=req.method, url=url,
                                    headers=headers,
                                    data=req.body,
                                    cert=client_cert,
                                    verify=verify_cert,
                                    timeout=60)
        except requests.ConnectionError:
            msg = _('The remote metadata server is temporarily unavailable. '
                    'Please try again later.')
            LOG.warning(msg)
            title = '503 Service Unavailable'
            return encode_http_reponse(title, title, msg)

        if resp.status_code == 200:
            return self._http_response(resp, req)
        if resp.status_code == 403:
            LOG.warning(
                'The remote metadata server responded with Forbidden. This '
                'response usually occurs when shared secrets do not match.'
            )
            # TODO(ralonsoh): add info in the returned HTTP message to the VM.
            return self._http_response(resp, req)
        if resp.status_code == 500:
            msg = _(
                'Remote metadata server experienced an internal server error.'
            )
            LOG.warning(msg)
            # TODO(ralonsoh): add info in the returned HTTP message to the VM.
            return self._http_response(resp, req)
        if resp.status_code in (400, 404, 409, 502, 503, 504):
            # TODO(ralonsoh): add info in the returned HTTP message to the VM.
            return self._http_response(resp, req)
        raise Exception(_('Unexpected response code: %s') % resp.status_code)

    def handle(self):
        try:
            request = self.request.recv(4096)
            LOG.debug('Request: %s', request.decode('utf-8'))
            f_request = io.BytesIO(request)
            req = webob.Request.from_file(f_request)
            instance_id, project_id = self._get_instance_and_project_id(req)
            if instance_id:
                res = self._proxy_request(instance_id, project_id, req)
                self.wfile.write(res)
                return

            network_id, router_id = self._get_instance_id(req)
            if network_id and router_id:
                title = '400 Bad Request'
                msg = (_('Both network %(network)s and router %(router)s '
                         'defined.') %
                         {'network': network_id, 'router': router_id})
                LOG.warning(msg)
            elif network_id:
                title = '404 Not Found'
                msg = _('Instance was not found on network %s.') % network_id
                LOG.warning(msg)
            else:
                title = '404 Not Found'
                msg = _('Instance was not found on router %s.') % router_id
                LOG.warning(msg)
            res = encode_http_reponse(title, title, msg)
            self.wfile.write(res)
        except Exception as exc:
            LOG.exception('Error while receiving data.')
            raise exc



# Copyright 2020 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from alembic import op
import sqlalchemy as sa


"""add dvr FIP gateway port network table

Revision ID: 2217c4222de6
Revises: Ibac91d24da2
Create Date: 2020-01-13 01:47:11.649472

"""

# revision identifiers, used by Alembic.
revision = '2217c4222de6'
down_revision = 'Ibac91d24da2'


def upgrade():
    op.create_table(
        'dvr_fip_gateway_port_network',
        sa.Column('network_id', sa.String(length=36),
                  sa.ForeignKey('networks.id', ondelete='CASCADE'),
                  primary_key=True),
        sa.Column('agent_id', sa.String(length=36),
                  sa.ForeignKey('agents.id', ondelete='CASCADE'),
                  primary_key=True)
    )



# Copyright 2024 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import attributes
from oslo_log import log

from neutron._i18n import _
from neutron.api import extensions
from neutron import manager
from neutron import service

LOG = log.getLogger(__name__)


def ovn_maintenance_worker():
    LOG.info('OVN maintenance process starting...')

    try:
        manager.init()
        ext_mgr = extensions.PluginAwareExtensionManager.get_instance()
        ext_mgr.extend_resources('2.0', attributes.RESOURCES)
        ovn_maintenance_worker = service.start_ovn_maintenance_worker()
        if not ovn_maintenance_worker:
            raise RuntimeError(_('OVN maintenance worker not loaded, ML2/OVN '
                                 'mechanism driver must be used'))
    except NotImplementedError:
        LOG.info('OVN maintenance worker was already started in parent '
                 'process by plugin.')
    else:
        ovn_maintenance_worker.wait()



# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.tests.unit.db import test_agentschedulers_db
from neutron.tests.unit.plugins.ml2 import test_plugin


class Ml2AgentSchedulerTestCase(
        test_agentschedulers_db.OvsAgentSchedulerTestCase):
    plugin_str = test_plugin.PLUGIN_NAME
    l3_plugin = ('neutron.services.l3_router.'
                 'l3_router_plugin.L3RouterPlugin')


class Ml2L3AgentNotifierTestCase(
        test_agentschedulers_db.OvsL3AgentNotifierTestCase):
    plugin_str = test_plugin.PLUGIN_NAME
    l3_plugin = ('neutron.services.l3_router.'
                 'l3_router_plugin.L3RouterPlugin')


class Ml2DhcpAgentNotifierTestCase(
        test_agentschedulers_db.OvsDhcpAgentNotifierTestCase):
    plugin_str = test_plugin.PLUGIN_NAME



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import netaddr
from neutron_lib.api.definitions import external_net
from neutron_lib.api.definitions import l3 as l3_apidef
from neutron_lib.api.definitions import portbindings
from neutron_lib.api.definitions import provider_net as pnet
from neutron_lib.api.definitions import qos_fip as qos_fip_apidef
from neutron_lib.api.definitions import qos_gateway_ip as qos_gateway_ip_apidef
from neutron_lib.callbacks import events
from neutron_lib.callbacks import priority_group
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as n_const
from neutron_lib import context as n_context
from neutron_lib.db import api as db_api
from neutron_lib.db import resource_extend
from neutron_lib import exceptions as n_exc
from neutron_lib.exceptions import availability_zone as az_exc
from neutron_lib.plugins import constants as plugin_constants
from neutron_lib.plugins import directory
from neutron_lib.services import base as service_base
from oslo_log import log

from neutron._i18n import _
from neutron.api.rpc.agentnotifiers import utils as notifier_utils
from neutron.common.ovn import constants as ovn_const
from neutron.common.ovn import extensions
from neutron.common.ovn import utils
from neutron.db.availability_zone import router as router_az_db
from neutron.db import dns_db
from neutron.db import extraroute_db
from neutron.db import l3_extra_gws_db
from neutron.db import l3_fip_pools_db
from neutron.db import l3_fip_port_details
from neutron.db import l3_fip_qos
from neutron.db import l3_gateway_ip_qos
from neutron.db import l3_gwmode_db
from neutron.db.models import l3 as l3_models
from neutron.plugins.ml2.drivers.ovn.mech_driver.ovsdb import ovn_client
from neutron.quota import resource_registry
from neutron.scheduler import l3_ovn_scheduler
from neutron.services.ovn_l3 import exceptions as ovn_l3_exc
from neutron.services.ovn_l3 import ovsdb_monitor
from neutron.services.ovn_l3.service_providers import driver_controller
from neutron.services.portforwarding.drivers.ovn import driver \
    as port_forwarding

LOG = log.getLogger(__name__)


@registry.has_registry_receivers
class OVNL3RouterPlugin(service_base.ServicePluginBase,
                        extraroute_db.ExtraRoute_dbonly_mixin,
                        l3_gwmode_db.L3_NAT_db_mixin,
                        dns_db.DNSDbMixin,
                        l3_fip_port_details.Fip_port_details_db_mixin,
                        router_az_db.RouterAvailabilityZoneMixin,
                        l3_fip_qos.FloatingQoSDbMixin,
                        l3_gateway_ip_qos.L3_gw_ip_qos_db_mixin,
                        l3_fip_pools_db.FloatingIPPoolsMixin,
                        l3_extra_gws_db.ExtraGatewaysDbOnlyMixin,
                        ):
    """Implementation of the OVN L3 Router Service Plugin.

    This class implements a L3 service plugin that provides
    router and floatingip resources and manages associated
    request/response.
    """

    # TODO(mjozefcz): Start consuming it from neutron-lib
    # once available.
    _supported_extension_aliases = (
        extensions.ML2_SUPPORTED_API_EXTENSIONS_OVN_L3)
    __filter_validation_support = True

    @resource_registry.tracked_resources(router=l3_models.Router,
                                         floatingip=l3_models.FloatingIP)
    def __init__(self):
        LOG.info("Starting OVNL3RouterPlugin")
        super().__init__()
        self._plugin_property = None
        self._mech = None
        self._initialize_plugin_driver()
        self._ovn_client_inst = None
        self.scheduler = l3_ovn_scheduler.get_scheduler()
        self.port_forwarding = port_forwarding.OVNPortForwarding(self)
        self.l3_driver_controller = driver_controller.DriverController(self)
        self.subscribe()
        self._l3_rpc_notifier = notifier_utils.RPCNotifierHandler()

    @property
    def l3_rpc_notifier(self):
        # The OVN L3 plugin notifier does not have an RPC instance. There is
        # no need to send any RPC update.
        return self._l3_rpc_notifier

    @staticmethod
    def _disable_qos_extensions_by_extension_drivers(aliases):
        qos_service_plugin = directory.get_plugin(plugin_constants.QOS)
        qos_fip_in_aliases = qos_fip_apidef.ALIAS in aliases
        qos_gwip_in_aliases = qos_gateway_ip_apidef.ALIAS in aliases
        if not qos_service_plugin and qos_fip_in_aliases:
            aliases.remove(qos_fip_apidef.ALIAS)
        if not qos_service_plugin and qos_gwip_in_aliases:
            aliases.remove(qos_gateway_ip_apidef.ALIAS)

    @property
    def supported_extension_aliases(self):
        if not hasattr(self, '_aliases'):
            self._aliases = self._supported_extension_aliases[:]
            self._disable_qos_extensions_by_extension_drivers(self._aliases)
        return self._aliases

    @property
    def _ovn_client(self):
        if self._ovn_client_inst is None:
            self._ovn_client_inst = ovn_client.OVNClient(self._nb_ovn,
                                                         self._sb_ovn)
        return self._ovn_client_inst

    @property
    def _nb_ovn(self):
        return self._plugin_driver.nb_ovn

    @property
    def _sb_ovn(self):
        return self._plugin_driver.sb_ovn

    @property
    def _plugin(self):
        if self._plugin_property is None:
            self._plugin_property = directory.get_plugin()
        return self._plugin_property

    def _initialize_plugin_driver(self):
        # This method initializes the mechanism driver variable and checks
        # if any of the valid drivers ('ovn', 'ovn-sync') is loaded.
        drivers = ('ovn', 'ovn-sync')
        for driver in drivers:
            try:
                self._mech = self._plugin.mechanism_manager.mech_drivers[
                    driver].obj
                break
            except KeyError:
                pass
        else:
            raise ovn_l3_exc.MechanismDriverNotFound(mechanism_drivers=drivers)

    @property
    def _plugin_driver(self):
        if self._mech is None:
            self._initialize_plugin_driver()
        return self._mech

    def get_plugin_type(self):
        return plugin_constants.L3

    def get_plugin_description(self):
        """returns string description of the plugin."""
        return ("L3 Router Service Plugin for basic L3 forwarding"
                " using OVN")

    def subscribe(self):
        # By default, the post fork initialization must be done first in the
        # ML2 plugin (the lower the priority number, the sooner is attended).
        registry.subscribe(self._post_fork_initialize,
                           resources.PROCESS, events.AFTER_INIT,
                           priority=priority_group.PRIORITY_DEFAULT + 1,
                           cancellable=True)

    def _post_fork_initialize(self, resource, event, trigger, payload=None):
        if not self._nb_ovn or not self._sb_ovn:
            raise ovn_l3_exc.MechanismDriverOVNNotReady()

        # Register needed events.
        self._nb_ovn.idl.notify_handler.watch_events([
            ovsdb_monitor.LogicalRouterPortEvent(self),
            ovsdb_monitor.LogicalRouterPortGatewayChassisEvent(self),
        ])

    def _add_neutron_router_interface(self, context, router_id,
                                      interface_info):
        try:
            router_interface_info = (
                super().add_router_interface(
                    context, router_id, interface_info))
        except n_exc.PortInUse:
            # NOTE(lucasagomes): If the port is already being used it means
            # the interface has been created already, let's just fetch it from
            # the database. Perhaps the code below should live in Neutron
            # itself, a get_router_interface() method in the main class
            # would be handy
            port = self._plugin.get_port(context, interface_info['port_id'])
            subnets = [self._plugin.get_subnet(context, s)
                       for s in utils.get_port_subnet_ids(port)]
            router_interface_info = (
                self._make_router_interface_info(
                    router_id, port['project_id'], port['id'],
                    port['network_id'], subnets[0]['id'],
                    [subnet['id'] for subnet in subnets]))

        return router_interface_info

    def create_floatingip(self, context, floatingip,
                          initial_status=n_const.FLOATINGIP_STATUS_DOWN):
        # The OVN L3 plugin creates floating IPs in down status by default,
        # whereas the L3 DB layer creates them in active status. So we keep
        # this method to create the floating IP in the DB with status down,
        # while the flavor drivers are responsible for calling the correct
        # backend to instatiate the floating IP in the data plane
        return super().create_floatingip(
            context, floatingip, initial_status)

    def update_floatingip_status(self, context, floatingip_id, status):
        fip = self.update_floatingip_status_retry(
            context, floatingip_id, status)
        registry.publish(
            resources.FLOATING_IP, events.AFTER_STATUS_UPDATE, self,
            payload=events.DBEventPayload(
                context, states=(fip,),
                resource_id=floatingip_id))
        return fip

    @db_api.retry_if_session_inactive()
    def update_floatingip_status_retry(self, context, floatingip_id, status):
        with db_api.CONTEXT_WRITER.using(context):
            return super().update_floatingip_status(
                context, floatingip_id, status)

    def _get_gateway_port_physnet_mapping(self):
        # This function returns all gateway ports with corresponding
        # external network's physnet
        net_physnet_dict = {}
        port_physnet_dict = {}
        l3plugin = directory.get_plugin(plugin_constants.L3)
        if not l3plugin:
            return port_physnet_dict
        context = n_context.get_admin_context()
        for net in l3plugin._plugin.get_networks(
                context, {external_net.EXTERNAL: [True]}):
            if net.get(pnet.NETWORK_TYPE) in [n_const.TYPE_FLAT,
                                              n_const.TYPE_VLAN]:
                net_physnet_dict[net['id']] = net.get(pnet.PHYSICAL_NETWORK)
        for port in l3plugin._plugin.get_ports(context, filters={
                'device_owner': [n_const.DEVICE_OWNER_ROUTER_GW]}):
            if utils.is_ovn_provider_router(
                    l3plugin.get_router(context, port['device_id'])):
                port_physnet_dict[port['id']] = net_physnet_dict.get(
                    port['network_id'])
        return port_physnet_dict

    def update_router_gateway_port_bindings(self, router, host):
        status = (n_const.PORT_STATUS_ACTIVE if host
                  else n_const.PORT_STATUS_DOWN)
        context = n_context.get_admin_context()
        filters = {'device_id': [router],
                   'device_owner': [n_const.DEVICE_OWNER_ROUTER_GW]}
        for port in self._plugin.get_ports(context, filters=filters):
            # FIXME(lucasagomes): Ideally here we would use only
            # one database transaction for the status and binding the
            # host but, even tho update_port_status() receives a "host"
            # parameter apparently it doesn't work for ports which the
            # device owner is router_gateway. We need to look into it and
            # fix the problem in Neutron before updating it here.
            if host:
                port = self._plugin.update_port(
                    context, port['id'],
                    {'port': {portbindings.HOST_ID: host}})
                # Updates OVN NB database with hostname for lsp router
                # gateway port
                with self._nb_ovn.transaction(check_error=True) as txn:
                    ext_ids = (
                        "external_ids",
                        {ovn_const.OVN_HOST_ID_EXT_ID_KEY: host},
                    )
                    txn.add(
                        self._nb_ovn.db_set(
                            "Logical_Switch_Port", port["id"], ext_ids
                        )
                    )
            if port['status'] != status:
                self._plugin.update_port_status(context, port['id'], status)

    def schedule_unhosted_gateways(self, event_from_chassis=None):
        # GW ports and its physnets.
        port_physnet_dict = self._get_gateway_port_physnet_mapping()
        # Filter out unwanted ports in case of event.
        if event_from_chassis:
            gw_chassis = self._nb_ovn.get_chassis_gateways(
                chassis_name=event_from_chassis)
            if not gw_chassis:
                return
            ports_impacted = []
            for gwc in gw_chassis:
                try:
                    ports_impacted.append(utils.get_port_id_from_gwc_row(gwc))
                except AttributeError:
                    # Malformed GWC format.
                    pass
            port_physnet_dict = {
                k: v
                for k, v in port_physnet_dict.items()
                if k in ports_impacted}
        if not port_physnet_dict:
            return
        # All chassis with physnets configured.
        chassis_with_physnets = self._sb_ovn.get_chassis_and_physnets()
        # All chassis with enable_as_gw_chassis set
        all_gw_chassis = self._sb_ovn.get_gateway_chassis_from_cms_options()
        chassis_with_azs = self._sb_ovn.get_chassis_and_azs()
        unhosted_gateways = self._nb_ovn.get_unhosted_gateways(
            port_physnet_dict, chassis_with_physnets,
            all_gw_chassis, chassis_with_azs)

        self._reschedule_lrps(unhosted_gateways)

    def _reschedule_lrps(self, lrps):
        # GW ports and its physnets.
        port_physnet_dict = self._get_gateway_port_physnet_mapping()
        # All chassis with physnets configured.
        chassis_with_physnets = self._sb_ovn.get_chassis_and_physnets()
        # All chassis with enable_as_gw_chassis set
        all_gw_chassis = self._sb_ovn.get_gateway_chassis_from_cms_options()
        chassis_with_azs = self._sb_ovn.get_chassis_and_azs()

        with self._nb_ovn.transaction(check_error=True) as txn:
            for g_name in lrps:
                # NOTE(fnordahl): Make scheduling decissions in ovsdbapp
                # command so that scheduling is done based on up to date
                # information as the transaction is applied.
                #
                # We pass in a reference to our class instance so that the
                # ovsdbapp command can call the scheduler methods from within
                # its context.
                txn.add(self._nb_ovn.schedule_unhosted_gateways(
                    g_name, self._sb_ovn, self, port_physnet_dict,
                    all_gw_chassis, chassis_with_physnets, chassis_with_azs))

    @staticmethod
    @registry.receives(resources.SUBNET, [events.AFTER_UPDATE])
    def _subnet_update(resource, event, trigger, payload):
        l3plugin = directory.get_plugin(plugin_constants.L3)
        if not l3plugin:
            return
        context = payload.context
        orig = payload.states[0]
        current = payload.latest_state
        orig_gw_ip = orig['gateway_ip']
        current_gw_ip = current['gateway_ip']
        if orig_gw_ip == current_gw_ip:
            return
        prefix = n_const.IP_ANY[netaddr.IPAddress(orig_gw_ip).version]
        gw_ports = l3plugin._plugin.get_ports(context, filters={
            'network_id': [orig['network_id']],
            'device_owner': [n_const.DEVICE_OWNER_ROUTER_GW],
            'fixed_ips': {'subnet_id': [orig['id']]},
        })
        router_ids = {port['device_id'] for port in gw_ports
                      if utils.is_ovn_provider_router(
                          l3plugin.get_router(context, port['device_id']))}
        remove = [{'destination': prefix, 'nexthop': orig_gw_ip}
                  ] if orig_gw_ip else []
        add = [{'destination': prefix, 'nexthop': current_gw_ip}
               ] if current_gw_ip else []
        add_columns = [{'external_ids': {
            ovn_const.OVN_ROUTER_IS_EXT_GW: 'true',
            ovn_const.OVN_SUBNET_EXT_ID_KEY: orig['id'],
            ovn_const.OVN_LRSR_EXT_ID_KEY: 'true'}}
        ]
        with l3plugin._nb_ovn.transaction(check_error=True) as txn:
            for router_id in router_ids:
                l3plugin._ovn_client.update_router_routes(
                    context, router_id, add, remove, add_columns, txn=txn)

    @staticmethod
    @registry.receives(
        resources.PORT,
        [events.BEFORE_UPDATE, events.AFTER_UPDATE]
    )
    def _port_update(resource, event, trigger, payload):
        l3plugin = directory.get_plugin(plugin_constants.L3)
        if not l3plugin:
            return

        context = payload.context
        current = payload.latest_state
        original = payload.states[0]

        # The OVN NB DB has a constraint where network has to be
        # greater than 0. Updating it with an empty network would
        # cause a constraint violation error. This problem happens
        # when the last IP of a LRP is deleted, in order to avoid it
        # an exception needs to be thrown before any write is performed
        # to the DB, since if not it would leave the Neutron DB and the
        # OVN DB unsync.
        # https://bugs.launchpad.net/neutron/+bug/1948457
        if (event == events.BEFORE_UPDATE and
                'fixed_ips' in current and not current['fixed_ips'] and
                utils.is_lsp_router_port(original) and
                utils.is_ovn_provider_router(
                    l3plugin.get_router(context, original['device_id']))):
            reason = _("Router port must have at least one IP.")
            raise n_exc.ServicePortInUse(port_id=original['id'], reason=reason)

    def get_router_availability_zones(self, router):
        lr = self._nb_ovn.get_lrouter(router['id'])
        if not lr:
            return []

        return [az.strip() for az in lr.external_ids.get(
            ovn_const.OVN_AZ_HINTS_EXT_ID_KEY, '').split(',')
                if az.strip()]

    def validate_availability_zones(self, context, resource_type,
                                    availability_zones):
        """Verify that the availability zones exist."""
        if not availability_zones or resource_type != 'router':
            return

        azs = {az['name'] for az in
               self._plugin_driver.list_availability_zones(context).values()}
        diff = set(availability_zones) - azs
        if diff:
            raise az_exc.AvailabilityZoneNotFound(
                availability_zone=', '.join(diff))

    @staticmethod
    @resource_extend.extends([l3_apidef.ROUTERS])
    def add_flavor_id(router_res, router_db):
        router_res['flavor_id'] = router_db['flavor_id']

    def update_router_gw_ports(self, context, network, subnet):
        self._update_router_gateway_ports(context, network, subnet)
